# -*- coding: utf-8 -*-

from __future__ import absolute_import

from .errors import NoSuchCubeError, NoSuchDimensionError, ModelError
from .common import read_json_file
from . import compat

__all__ = [
    "Namespace",
]

class Namespace(object):
    def __init__(self, name=None, parent=None):
        """Creates a cubes namespace – an object that provides model objects
        from the providers."""
        # TODO: Assign this on __init__, namespaces should not be freely
        # floating until anchored in another namespace
        self.parent = parent
        self.name = name
        self.namespaces = {}
        self.providers = []
        self.translations = {}

    def namespace(self, path, create=False):
        """Returns a tuple (`namespace`, `remainder`) where `namespace` is
        the deepest namespace in the namespace hierarchy and `remainder` is
        the remaining part of the path that has no namespace (is an object
        name or contains part of external namespace).

        If path is empty or not provided then returns self.

        If `create` is `True` then the deepest namespace is created if it does
        not exist.
        """

        if not path:
            return (self, None)

        if isinstance(path, compat.string_type):
            path = path.split(".")

        namespace = self
        for i, element in enumerate(path):
            remainder = path[i+1:]
            if element in namespace.namespaces:
                namespace = namespace.namespaces[element]
                found = True
            else:
                remainder = path[i:]
                break

        if not create:
            return (namespace, ".".join(remainder) or None)
        else:
            for element in remainder:
                namespace = namespace.create_namespace(element)

            return (namespace, None)

    def create_namespace(self, name):
        """Create a namespace `name` in the receiver."""
        if self.name:
            nsname = "%s.%s" % (self.name, name)
        else:
            nsname = name

        namespace = Namespace(nsname, parent=self)
        self.namespaces[name] = namespace

        return namespace

    def find_cube(self, cube_ref):
        """Returns a tuple (`namespace`, `provider`, `basename`) where
        `namespace` is a namespace conaining `cube`, `provider` providers the
        model for the cube and `basename` is a name of the `cube` within the
        `namespace`. For example: if cube is ``slicer.nested.cube`` and there
        is namespace ``slicer`` then that namespace is returned and the
        `basename` will be ``nested.cube``.

        Raises `NoSuchCubeError` when there is no cube with given reference.
        """

        cube_ref = str(cube_ref)
        split = cube_ref.split(".")
        if len(split) > 1:
            path = split[0:-1]
            cube_ref = split[-1]
        else:
            path = []
            cube_ref = cube_ref

        (namespace, remainder) = self.namespace(path)

        if remainder:
            basename = "{}.{}".format(remainder, cube_ref)
        else:
            basename = cube_ref

        # Find first provider that knows about the cube `name`
        provider = None

        for provider in namespace.providers:
            if provider.has_cube(cube_ref):
                break
        else:
            provider = None

        if not provider:
            raise NoSuchCubeError("Unknown cube '{}'".format(cube_ref),
                                  cube_ref)

        return (namespace, provider, basename)


    def list_cubes(self, recursive=False):
        """Retursn a list of cube info dictionaries with keys: `name`,
        `label`, `description`, `category` and `info`."""

        all_cubes = []
        cube_names = set()
        for provider in self.providers:
            cubes = provider.list_cubes()
            # Cehck for duplicity
            for cube in cubes:
                name = cube["name"]
                if name in cube_names:
                    raise ModelError("Duplicate cube '%s'" % name)
                cube_names.add(name)

            all_cubes += cubes

        if recursive:
            for name, ns in self.namespaces.items():
                cubes = ns.list_cubes(recursive=True)
                for cube in cubes:
                    cube["name"] = "%s.%s" % (name, cube["name"])
                all_cubes += cubes

        return all_cubes

    # TODO: change to find_dimension() analogous to the find_cube(). Let the
    # caller to perform actual dimension creation using the provider
    def dimension(self, name, locale=None, templates=None, local_only=False):
        dim = None

        for provider in self.providers:
            # TODO: use locale
            try:
                dim = provider.dimension(name, locale=locale,
                                         templates=templates)
            except NoSuchDimensionError:
                pass
            else:
                return dim

        # If we are not looking for dimension within this namespace only,
        # traverse the namespace hierarchy, if there is one
        if not local_only and self.parent:
            return self.parent.dimension(name, locale, templates)

        raise NoSuchDimensionError("Unknown dimension '%s'" % str(name), name)

    def add_provider(self, provider):
        self.providers.append(provider)

    def add_translation(self, lang, translation):
        """Registers and merges `translation` for language `lang`"""
        try:
            trans = self.translations[lang]
        except KeyError:
            trans = {}
            self.translations[lang] = trans

        # Is it a path?
        if isinstance(translation, compat.string_type):
            translation = read_json_file(translation)

        trans.update(translation)

    def translation_lookup(self, lang):
        """Returns translation in language `lang` for model object `obj`
        within `context` (cubes, dimensions, attributes, ...).  Looks in
        parent if current namespace does not have the translation."""

        lookup = []
        visited = set()

        # Find namespaces with translation language
        ns = self
        while ns and ns not in visited:
            if lang in ns.translations:
                lookup.append(ns.translations[lang])
            visited.add(ns)
            ns = ns.parent

        return lookup




# -*- coding=utf -*-
"""
    cubes.datastructures
    ~~~~~~~~~~~~~~~~~~~~~

    Utility data structures.
"""


from collections import MutableMapping
import sys


__all__ = [
    "AttributeGetter",
    "AttributeDictMixin",
    "AttributeDict",
    "DictAttribute",
    "FlatAccessDict",
]


# Helper class for Python attribute getter
class AttributeGetter(object):
    def __init__(self, getter):
        self.getter = getter

    def __getattr__(self, name):
        return self.getter(name)

#
# Credits:
# Originally from the Celery project:  http://www.celeryproject.org
#
class AttributeDictMixin(object):
    """Augment classes with a Mapping interface by adding attribute access.

    I.e. `d.key -> d[key]`.

    """

    def __getattr__(self, k):
        """`d.key -> d[key]`"""
        try:
            return self[k]
        except KeyError:
            raise AttributeError(
                '{0!r} object has no attribute {1!r}'.format(
                    type(self).__name__, k))

    def __setattr__(self, key, value):
        """`d[key] = value -> d.key = value`"""
        self[key] = value


class AttributeDict(dict, AttributeDictMixin):
    """Dict subclass with attribute access."""
    pass


class DictAttribute(object):
    """Dict interface to attributes.

    `obj[k] -> obj.k`
    `obj[k] = val -> obj.k = val`

    """
    obj = None

    def __init__(self, obj):
        object.__setattr__(self, 'obj', obj)

    def __getattr__(self, key):
        return getattr(self.obj, key)

    def __setattr__(self, key, value):
        return setattr(self.obj, key, value)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def setdefault(self, key, default):
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return default

    def __getitem__(self, key):
        try:
            return getattr(self.obj, key)
        except AttributeError:
            raise KeyError(key)

    def __setitem__(self, key, value):
        setattr(self.obj, key, value)

    def __contains__(self, key):
        return hasattr(self.obj, key)

    def _iterate_keys(self):
        return iter(dir(self.obj))
    iterkeys = _iterate_keys

    def __iter__(self):
        return self._iterate_keys()

    def _iterate_items(self):
        for key in self._iterate_keys():
            yield key, getattr(self.obj, key)
    iteritems = _iterate_items

    def _iterate_values(self):
        for key in self._iterate_keys():
            yield getattr(self.obj, key)
    itervalues = _iterate_values

    if sys.version_info[0] == 3:  # pragma: no cover
        items = _iterate_items
        keys = _iterate_keys
        values = _iterate_values
    else:

        def keys(self):
            return list(self)

        def items(self):
            return list(self._iterate_items())

        def values(self):
            return list(self._iterate_values())

MutableMapping.register(DictAttribute)

class FlatAccessDict(dict):
    """A dictionary where items can be accessed by a 'dotted' path. For
    example `d["foo.bar"]` will be the same as `d["foo"]["bar"]`"""

    def __getitem__(self, key):
        path = key.split(".")
        obj = super(FlatAccessDict, self).__getitem__(path[0])

        for item in path[1:]:
            obj = obj[item]

        return obj

    def __contains__(self, key):
        path = key.split(".")

        if not super(FlatAccessDict, self).__contains__(path[0]):
            return False

        obj = self[path[0]]

        for item in path[1:]:
            if not item in obj:
                return False
            obj = obj[item]

        return True

    def pop(self, key, default=None):
        path = key.split(".")
        owner = self

        # TODO: should we `del` when owner of popped object is empty?

        for item in path[:-1]:
            owner = owner[item]

        if owner is self:
            return super(FlatAccessDict, self).pop(key, default)
        else:
            return owner.pop(path[-1], default)




# -*- coding: utf-8 -*-

import os
from .base import create_server
from .utils import str_to_bool

# Set the configuration file
try:
    CONFIG_PATH = os.environ["SLICER_CONFIG"]
except KeyError:
    CONFIG_PATH = os.path.join(os.getcwd(), "slicer.ini")

application = create_server(CONFIG_PATH)

debug = os.environ.get("SLICER_DEBUG")
if debug and str_to_bool(debug):
    application.debug = True



# -*- encoding: utf-8 -*-

from __future__ import absolute_import

import itertools

from .errors import ArgumentError

__all__ = [
    "combined_cuboids",
    "combined_levels",
    "hierarchical_cuboids"
]

def combined_cuboids(dimensions, required=None):
    """Returns a list of all combinations of `dimensions` as tuples. For
    example, if `dimensions` is: ``['date', 'product']`` then it returns:

        ``[['date', 'cpv'], ['date'], ['cpv']]``
    """

    required = tuple(required) if required else ()

    for dim in required:
        if dim not in dimensions:
            raise ArgumentError("Required dimension '%s' is not in list of "
                                "dimensions to be combined." % str(dim))

    cuboids = []
    to_combine = [dim for dim in dimensions if not dim in required]

    for i in range(len(to_combine), 0, -1):
        combos = itertools.combinations(to_combine, i)
        combos = [required+combo for combo in combos]

        cuboids += tuple(combos)

    if required:
        cuboids = [required] + cuboids

    return cuboids

def combined_levels(dimensions, default_only=False):
    """Create a cartesian product of levels from all `dimensions`. For
    example, if dimensions are _date_, _product_ then result will be:
    levels of _date_ X levels of _product_. Each element of the returned list
    is a list of tuples (`dimension`, `level`)
    """
    groups = []
    for dim in dimensions:
        if default_only:
            levels = dim.hierarchy().levels
        else:
            levels = dim.levels

        group = [(str(dim), str(level)) for level in levels]
        groups.append(group)

    return tuple(itertools.product(*groups))


def hierarchical_cuboids(dimensions, required=None, default_only=False):
    """Returns a list of cuboids with all hierarchical level combinations."""
    cuboids = combined_cuboids(dimensions, required)

    result = []
    for cuboid in cuboids:
        result += list(combined_levels(cuboid, default_only))

    return result




# -*- coding: utf-8 -*-

from __future__ import absolute_import
from . import compat
# Global Context – top level namespace and objects in other namespaces
# Local Context - within object being translated

# TODO: work in progress
# TODO: rename to something nicer
# TODO: reuse the lookup function
# TODO: Change this to use translation lookup instead of just one translation

class ModelObjectLocalizationContext(object):
    def __init__(self, translation, context, object_type, object_name):
        self.translation = translation
        self.object_type = object_type
        self.object_name = object_name
        self.context = context

    def get(self, key, default=None):
        try:
            return self.translation[key]
        except KeyError:
            return self.context.get(self.object_type, self.object_name, key,
                                    default)

    def object_localization(self, object_type, name):
        try:
            objects = self.translation[object_type]
        except KeyError:
            objects = self.context.translation.get(object_type, {})

        try:
            trans = objects[name]
        except KeyError:
            return ModelObjectLocalizationContext({}, self.context,
                                              object_type, name)

        # Make string-only translations as translations of labels
        if isinstance(trans, compat.string_type):
            trans = {"label": trans}

        return ModelObjectLocalizationContext(trans, self.context,
                                              object_type, name)

class LocalizationContext(object):
    def __init__(self, translation, parent=None):
        self.translation = translation
        self.parent = parent

    def object_localization(self, object_type, name):
        try:
            objects = self.translation[object_type]
        except KeyError:
            return ModelObjectLocalizationContext({}, self, object_type, name)

        try:
            trans = objects[name]
        except KeyError:
            return ModelObjectLocalizationContext({}, self, object_type, name)

        # Make string-only translations as translations of labels
        if isinstance(trans, compat.string_type):
            trans = {"label": trans}

        return ModelObjectLocalizationContext(trans, self, object_type, name)

    def get(self, object_type, object_name, key, default=None):
        try:
            objects = self.translation[object_type]
        except KeyError:
            return default

        try:
            trans = objects[object_name]
        except KeyError:
            return default

        # Accept plain label translation – string only, no dictionary (similar
        # as above)
        if isinstance(trans, compat.string_type):
            if key == "label":
                return trans
            else:
                return default

        return trans.get(key, default)

    def _get_translation(self, obj, type_):
        """Returns translation in language `lang` for model object `obj` of
        type `type_`. The type can be: ``cube`` or ``dimension``. Looks in
        parent if current namespace does not have the translation."""

        lookup = []
        visited = set()

        # Find namespaces with translation language
        ns = self.namespace
        while ns and ns not in visited:
            if self.lang in ns.translations:
                lookup.append(ns.translation[lang])
            visited.add(ns)
            ns = ns.parent

        lookup_map = {
            "cube": "cubes",
            "dimension": "dimensions",
            "defaults": "defaults"
        }

        objkey = lookup_map[type_]

        for trans in lookup:
            if objkey in trans and obj in trans[objkey]:
                return trans[objkey][obj]

        return None




# -*- encoding=utf -*-
"""SQL Browser"""

from __future__ import absolute_import

import collections

try:
    import sqlalchemy
    import sqlalchemy.sql as sql

except ImportError:
    from ...common import MissingPackage
    sqlalchemy = sql = MissingPackage("sqlalchemy", "SQL aggregation browser")

from ..statutils import available_calculators
from ..browser import AggregationBrowser, AggregationResult, Drilldown
from ..logging import get_logger
from ..errors import ArgumentError, InternalError
from ..stores import Store
from ..cells import Cell, PointCut
from ..model import collect_attributes
from .. import compat

from .functions import available_aggregate_functions
from .mapper import DenormalizedMapper, StarSchemaMapper, map_base_attributes
from .mapper import distill_naming
from .query import StarSchema, QueryContext, to_join, FACT_KEY_LABEL
from .utils import paginate_query, order_query


__all__ = [
    "SQLBrowser",
]


class SQLBrowser(AggregationBrowser):
    """SnowflakeBrowser is a SQL-based AggregationBrowser implementation that
    can aggregate star and snowflake schemas without need of having
    explicit view or physical denormalized table.

    Attributes:

    * `cube` - browsed cube
    * `store` - a `Store` object or a SQLAlchemy engine
    * `locale` - locale used for browsing
    * `debug` - output SQL to the logger at INFO level

    Other options in `kwargs`:
    * `metadata` – SQLAlchemy metadata, if `store` is an engine or a
       connection (not a `Store` object)
    * `tables` – tables and/or table expressions used in the star schema
      (refer to the :class:`StarSchema` for more information)
    * `options` - passed to the mapper

    Tuning:

    * `include_summary` - it ``True`` then summary is included in
      aggregation result. Turned on by default.
    * `include_cell_count` – if ``True`` then total cell count is included
      in aggregation result. Turned on by default.
      performance reasons
    * `safe_labels` – safe labelling of the attributes in databases which
      don't allow characters such as ``.`` dots in column names

    Limitations:

    * only one locale can be used for browsing at a time
    * locale is implemented as denormalized: one column for each language

    """

    __description__ = """
    SQL - relational database browser (for ROLAP). Generates statements on top
    of star or snowflake schemas.
    """

    __options__ = [
        {
            "name": "include_summary",
            "description": "Include aggregation summary "\
                           "(requires extra statement)",
            "type": "bool"
        },
        {
            "name": "include_cell_count",
            "type": "bool"
        },
        {
            "name": "use_denormalization",
            "type": "bool"
        },
        {
            "name": "safe_labels",
            "description": "Use internally SQL statement column labels " \
                           "without special characters",
            "type": "bool"
        }

    ]

    def __init__(self, cube, store, locale=None, debug=False, **kwargs):
        """Create a SQL Browser."""

        super(SQLBrowser, self).__init__(cube, store)

        if not cube:
            raise ArgumentError("Cube for browser should not be None.")

        self.logger = get_logger()

        self.cube = cube
        self.locale = locale or cube.locale
        self.debug = debug

        # Database connection and metadata
        # --------------------------------

        if isinstance(store, Store):
            self.connectable = store.connectable
            metadata = store.metadata
        else:
            self.connectable = store

            metadata = kwargs.get("metadata",
                                  sqlalchemy.MetaData(bind=self.connectable))

        # Options
        # -------

        # Merge options with store options
        options = {}
        options.update(store.options)
        options.update(kwargs)

        self.include_summary = options.get("include_summary", True)
        self.include_cell_count = options.get("include_cell_count", True)

        self.safe_labels = options.get("safe_labels", False)
        if self.safe_labels:
            self.logger.debug("using safe labels for cube {}"
                              .format(cube.name))

        # Whether to ignore cells where at least one aggregate is NULL
        # TODO: this is undocumented
        self.exclude_null_agregates = options.get("exclude_null_agregates",
                                                  False)

        # Mapper
        # ------

        # Mapper is responsible for finding corresponding physical columns to
        # dimension attributes and fact measures. It also provides information
        # about relevant joins to be able to retrieve certain attributes.

        if options.get("is_denormalized", options.get("use_denormalization")):
            mapper = DenormalizedMapper
        else:
            mapper = StarSchemaMapper

        self.logger.debug("using mapper %s for cube '%s' (locale: %s)" %
                          (str(mapper.__name__), cube.name, locale))

        # Prepare the mappings of base attributes
        #
        naming = distill_naming(options)
        (fact_name, mappings) = map_base_attributes(cube, mapper,
                                                    naming=naming,
                                                    locale=locale)

        tables = options.get("tables")

        # Prepare Join objects
        if cube.joins:
            joins = [to_join(join) for join in cube.joins]
        else:
            joins = []

        self.star = StarSchema(self.cube.name,
                               metadata,
                               mappings=mappings,
                               fact=fact_name,
                               joins=joins,
                               schema=naming.schema,
                               tables=tables)

        # Extract hierarchies
        # -------------------
        #
        self.hierarchies = self.cube.distilled_hierarchies

    def features(self):
        """Return SQL features. Currently they are all the same for every
        cube, however in the future they might depend on the SQL engine or
        other factors."""

        features = {
            "actions": ["aggregate", "fact", "members", "facts", "cell"],
            "aggregate_functions": available_aggregate_functions(),
            "post_aggregate_functions": available_calculators()
        }

        return features

    def is_builtin_function(self, funcname):
        """Returns `True` if the function `funcname` is backend's built-in
        function."""

        return funcname in available_aggregate_functions()

    def fact(self, key_value, fields=None):
        """Get a single fact with key `key_value` from cube.

        Number of SQL queries: 1."""

        (statement, labels) = self.denormalized_statement(attributes=fields,
                                                          include_fact_key=True)
        condition = statement.columns[FACT_KEY_LABEL] == key_value
        statement = statement.where(condition)

        cursor = self.execute(statement, "fact")
        row = cursor.fetchone()

        if row:
            # Convert SQLAlchemy object into a dictionary
            record = dict(zip(labels, row))
        else:
            record = None

        cursor.close()

        return record

    def facts(self, cell=None, fields=None, order=None, page=None,
              page_size=None, fact_list=None):
        """Return all facts from `cell`, might be ordered and paginated.

        `fact_list` is a list of fact keys to be selected. Might be used to
        fetch multiple facts using single query instead of multiple `fact()`
        queries.

        Number of SQL queries: 1.
        """
        attrs = self.cube.get_attributes(fields)
        cell = cell or Cell(self.cube)

        (statement, labels) = self.denormalized_statement(cell=cell,
                                                          attributes=attrs,
                                                          include_fact_key=True)

        if fact_list is not None:
            in_condition = self.star.fact_key_column.in_(fact_list)
            statement = statement.where(in_condition)

        statement = paginate_query(statement, page, page_size)

        # TODO: use natural order
        statement = order_query(statement,
                                order,
                                natural_order={},
                                labels=labels)

        cursor = self.execute(statement, "facts")

        return ResultIterator(cursor, labels)

    def test(self, aggregate=False):
        """Tests whether the statement can be constructed and executed. Does
        not return anything, but raises an exception if there are issues with
        the generated statements. By default it tests only denormalized
        statement by fetching one row. If `aggregate` is `True` then test also
        aggregation."""
        (statement, _) = self.denormalized_statement()
        statement = statement.limit(1)
        result = self.connectable.execute(statement)
        result.close()

        aggs = self.cube.all_aggregate_attributes
        dd = Drilldown()

        (statement, labels) = self.aggregation_statement(aggregates=aggs,
                                                         cell=Cell(self.cube),
                                                         drilldown=dd,
                                                         for_summary=True)
        result = self.connectable.execute(statement)
        result.close()

    def provide_members(self, cell, dimension, depth=None, hierarchy=None,
                        levels=None, attributes=None, page=None,
                        page_size=None, order=None):
        """Return values for `dimension` with level depth `depth`. If `depth`
        is ``None``, all levels are returned.

        Number of database queries: 1.
        """
        if not attributes:
            attributes = []
            for level in levels:
                attributes += level.attributes

        (statement, labels) = self.denormalized_statement(attributes, cell)
        # Order and paginate
        #
        statement = statement.group_by(*statement.columns)
        statement = order_query(statement,
                                order,
                                labels=labels)
        statement = paginate_query(statement, page, page_size)

        result = self.execute(statement, "members")

        return ResultIterator(result, labels)

    def path_details(self, dimension, path, hierarchy=None):
        """Returns details for `path` in `dimension`. Can be used for
        multi-dimensional "breadcrumbs" in a used interface.

        Number of SQL queries: 1.
        """
        dimension = self.cube.dimension(dimension)
        hierarchy = dimension.hierarchy(hierarchy)

        cut = PointCut(dimension, path, hierarchy=hierarchy)
        cell = Cell(self.cube, [cut])

        attributes = []
        for level in hierarchy.levels[0:len(path)]:
            attributes += level.attributes

        (statement, labels) = self.denormalized_statement(attributes,
                                                          cell,
                                                          include_fact_key=True)
        statement = statement.limit(1)
        cursor = self.execute(statement, "path details")

        row = cursor.fetchone()

        if row:
            member = dict(zip(labels, row))
        else:
            member = None

        return member

    def execute(self, statement, label=None):
        """Execute the `statement`, optionally log it. Returns the result
        cursor."""
        self._log_statement(statement, label)
        return self.connectable.execute(statement)

    def provide_aggregate(self, cell, aggregates, drilldown, split, order,
                          page, page_size, **options):
        """Return aggregated result.

        Arguments:

        * `cell`: cell to be aggregated
        * `measures`: aggregates of these measures will be considered
        * `aggregates`: aggregates to be considered
        * `drilldown`: list of dimensions or list of tuples: (`dimension`,
          `hierarchy`, `level`)
        * `split`: an optional cell that becomes an extra drilldown segmenting
          the data into those within split cell and those not within
        * `attributes`: list of attributes from drilled-down dimensions to be
          returned in the result
        * `across`: list of other cubes to be drilled across

        Query tuning:

        * `include_cell_count`: if ``True`` (``True`` is default) then
          `result.total_cell_count` is
          computed as well, otherwise it will be ``None``.
        * `include_summary`: if ``True`` (default) then summary is computed,
          otherwise it will be ``None``

        Result is paginated by `page_size` and ordered by `order`.

        Number of database queries:

        * without drill-down: 1 – summary
        * with drill-down (default): 3 – summary, drilldown, total drill-down
          record count

        Notes:

        * measures can be only in the fact table

        """

        # TODO: implement reminder

        result = AggregationResult(cell=cell, aggregates=aggregates,
                                   drilldown=drilldown,
                                   has_split=split is not None)

        # Summary
        # -------

        if self.include_summary or not (drilldown or split):
            (statement, labels) = self.aggregation_statement(cell,
                                                             aggregates=aggregates,
                                                             drilldown=drilldown,
                                                             for_summary=True)

            cursor = self.execute(statement, "aggregation summary")
            row = cursor.first()

            if row:
                # Convert SQLAlchemy object into a dictionary
                record = dict(zip(labels, row))
            else:
                record = None

            result.summary = record

        # Drill-down
        # ----------
        #
        # Note that a split cell if present prepends the drilldown

        if drilldown or split:
            if not (page_size and page is not None):
                self.assert_low_cardinality(cell, drilldown)

            result.levels = drilldown.result_levels(include_split=bool(split))
            natural_order = drilldown.natural_order

            self.logger.debug("preparing drilldown statement")

            (statement, labels) = self.aggregation_statement(cell,
                                                             aggregates=aggregates,
                                                             drilldown=drilldown,
                                                             split=split)
            # Get the total cell count before the pagination
            #
            if self.include_cell_count:
                count_statement = statement.alias().count()
                counts = self.execute(count_statement)
                result.total_cell_count = counts.scalar()

            # Order and paginate
            #
            statement = order_query(statement,
                                    order,
                                    natural_order,
                                    labels=labels)
            statement = paginate_query(statement, page, page_size)

            cursor = self.execute(statement, "aggregation drilldown")

            result.cells = ResultIterator(cursor, labels)
            result.labels = labels

        # If exclude_null_aggregates is True then don't include cells where
        # at least one of the bult-in aggregates is NULL
        if result.cells is not None and self.exclude_null_agregates:
            native_aggs = [agg.ref for agg in aggregates
                           if agg.function and self.is_builtin_function(agg.function)]
            result.exclude_if_null = native_aggs

        return result

    def _create_context(self, attributes):
        """Create a query context for `attributes`. The `attributes` should
        contain all attributes that will be somehow involved in the query."""

        collected = self.cube.collect_dependencies(attributes)
        return QueryContext(self.star,
                            attributes=collected,
                            hierarchies=self.hierarchies,
                            parameters=None,
                            safe_labels=self.safe_labels)

    def denormalized_statement(self, attributes=None, cell=None,
                               include_fact_key=False):
        """Returns a tuple (`statement`, `labels`) representing denormalized
        star statement restricted by `cell`. If `attributes` is not specified,
        then all cube's attributes are selected. The returned `labels` are
        correct labels to be applied to the iterated result in case of
        `safe_labels`."""

        attributes = attributes or self.cube.all_fact_attributes

        refs = [attr.ref for attr in collect_attributes(attributes, cell)]
        context_attributes = self.cube.get_attributes(refs)
        context = self._create_context(context_attributes)

        if include_fact_key:
            selection = [self.star.fact_key_column]
        else:
            selection = []

        names = [attr.ref for attr in attributes]
        selection += context.get_columns(names)

        cell_condition = context.condition_for_cell(cell)

        statement = sql.expression.select(selection,
                                          from_obj=context.star,
                                          whereclause=cell_condition)

        return (statement, context.get_labels(statement.columns))

    # Aggregate
    # =========
    #
    # This is the reason of our whole existence.
    #
    def aggregation_statement(self, cell, aggregates, drilldown=None,
                              split=None, for_summary=False):
        """Builds a statement to aggregate the `cell` and reutrns a tuple
        (`statement`, `labels`). `statement` is a SQLAlchemy statement object,
        `labels` is a list of attribute names selected in the statement. The
        `labels` should be applied on top of the result iterator, since the
        real columns might have simplified labels when `safe_labels` is
        ``True``.

        * `cell` – `Cell` to aggregate
        * `aggregates` – list of aggregates to consider (should not be empty)
        * `drilldown` – an optional `Drilldown` object
        * `split` – split cell for split condition
        * `for_summary` – do not perform `GROUP BY` for the drilldown. The
          drilldown is used only for choosing tables to join
        """
        # * `across` – cubes that share dimensions

        # TODO: PTD

        # Basic assertions

        if not aggregates:
            raise ArgumentError("List of aggregates should not be empty")

        if not isinstance(drilldown, Drilldown):
            raise InternalError("Drilldown should be a Drilldown object. "
                                "Is '{}'".format(type(drilldown)))

        # 1. Gather attributes
        #

        self.logger.debug("prepare aggregation statement. cell: '%s' "
                          "drilldown: '%s' for summary: %s" %
                          (",".join([compat.to_unicode(cut) for cut in cell.cuts]),
                           drilldown, for_summary))

        # TODO: it is verylikely that the _create_context is not getting all
        # attributes, for example those that aggregate depends on
        refs = collect_attributes(aggregates, cell, drilldown, split)
        attributes = self.cube.get_attributes(refs, aggregated=True)
        context = self._create_context(attributes)

        # Drilldown – Group-by
        # --------------------
        #
        # SELECT – Prepare the master selection
        #     * master drilldown items

        selection = context.get_columns([attr.ref for attr in
                                         drilldown.all_attributes])

        # SPLIT
        # -----
        if split:
            selection.append(context.column_for_split(split))

        # WHERE
        # -----
        condition = context.condition_for_cell(cell)

        group_by = selection[:] if not for_summary else None

        # TODO: coalesce if there are outer joins
        # TODO: ignore post-aggregations
        aggregate_cols = context.get_columns([agg.ref for agg in aggregates])

        if for_summary:
            # Don't include the group-by part (see issue #157 for more
            # information)
            selection = aggregate_cols
        else:
            selection += aggregate_cols

        statement = sql.expression.select(selection,
                                          from_obj=context.star,
                                          use_labels=True,
                                          whereclause=condition,
                                          group_by=group_by)

        return (statement, context.get_labels(statement.columns))

    def _log_statement(self, statement, label=None):
        label = "SQL(%s):" % label if label else "SQL:"
        self.logger.debug("%s\n%s\n" % (label, str(statement)))


class ResultIterator(object):
    """
    Iterator that returns SQLAlchemy ResultProxy rows as dictionaries
    """
    def __init__(self, result, labels):
        self.result = result
        self.batch = None
        self.labels = labels
        self.exclude_if_null = None

    def __iter__(self):
        while True:
            if not self.batch:
                many = self.result.fetchmany()
                if not many:
                    break
                self.batch = collections.deque(many)

            row = self.batch.popleft()

            if self.exclude_if_null \
                    and any(row[agg] is None for agg in self.exclude_if_null):
                continue

            yield dict(zip(self.labels, row))



# -*- coding: utf-8 -*-

from collections import deque
from functools import partial
from math import sqrt

from .errors import ArgumentError, InternalError, ModelError
from . import compat

__all__ = [
    "CALCULATED_AGGREGATIONS",
    "calculators_for_aggregates",
    "available_calculators",
    "aggregate_calculator_labels"
]


def calculators_for_aggregates(cube, aggregates, drilldown_levels=None,
                               split=None):
    """Returns a list of calculator function objects that implements
    aggregations by calculating on retrieved results, given a particular
    drilldown. Only post-aggregation calculators are returned.

    Might return an empty list if there is no post-aggregation witin
    aggregate functions.

    Aggregates are supposed to be only those aggregates which refer to a
    post-aggregation function.
    """
    # If we have an aggregation function, then we consider the aggregate
    # already processed
    functions = []

    for aggregate in aggregates:
        # Ignore function if the backend already handles it
        try:
            factory = CALCULATED_AGGREGATIONS[aggregate.function]
        except KeyError:
            raise ArgumentError("Unknown post-calculation function '%s' for "
                                "aggregate '%s'" % (aggregate.function,
                                                    aggregate.name))

        if aggregate.measure:
            source = cube.measure_aggregate(aggregate.measure)
        else:
            raise InternalError("No measure specified for aggregate '%s' in "
                                "cube '%s'" % (aggregate.name, cube.name))

        func = factory(aggregate, source.ref, drilldown_levels, split)
        functions.append(func)

    return functions

def weighted_moving_average(values):
    n = len(values)
    denom = n * (n + 1) / 2
    total = 0.0
    idx = 1
    for val in values:
        total += float(idx) * float(val)
        idx += 1
    return round(total / denom, 4)


def simple_moving_average(values):
    # use all the values
    return round(compat.reduce(lambda i, c: float(c) + i, values, 0.0) / len(values), 2)

def simple_moving_sum(values):
    return compat.reduce(lambda i, c: i + c, values, 0)


def _variance(values):
    n, mean, std = len(values), 0, 0
    for a in values:
        mean = mean + a
    mean = mean / float(n)
    if n < 2:
        return mean, 0
    for a in values:
        std = std + (a - mean)**2
    return mean, (std / float(n-1))

def simple_relative_stdev(values):
    mean, var = _variance(values)
    return round(((sqrt(var)/mean) if mean > 0 else 0), 4)

def simple_variance(values):
    mean, var = _variance(values)
    return round(var, 2)

def simple_stdev(values):
    mean, var = _variance(values)
    return round(sqrt(var), 2)

def _window_function_factory(aggregate, source, drilldown_paths, split_cell, window_function, label):
    """Returns a moving average window function. `aggregate` is the target
    aggergate. `window_function` is concrete window function."""

    # If the level we're drilling to doesn't have aggregation_units configured,
    # we're not doing any calculations

    key_drilldown_paths = []
    window_size = None
    drilldown_paths = drilldown_paths or []

    if aggregate.window_size:
        window_size = aggregate.window_size
    else:
        # TODO: this is the old depreciated way, remove when not needed
        for path in drilldown_paths:
            relevant_level = path.levels[-1]
            these_num_units = None
            if relevant_level.info:
                these_num_units = relevant_level.info.get('aggregation_units', None)
            if these_num_units is None:
                key_drilldown_paths.append(path)
            else:
                window_size = these_num_units

    if window_size is None:
        window_size = 1

    elif not isinstance(window_size, int) or window_size < 1:
        raise ModelError("window size for aggregate '%s' sohuld be an integer "
                         "greater than or equeal 1" % aggregate.name)

    # Create a composite key for grouping:
    #   * split dimension, if used
    #   * key from drilldown path levels
    #
    # If no key_drilldown_paths, the key is always the empty tuple.

    window_key = []
    if split_cell:
        from .browser import SPLIT_DIMENSION_NAME
        window_key.append(SPLIT_DIMENSION_NAME)
    for dditem in key_drilldown_paths:
        window_key += [level.key.ref for level in dditem.levels]

    # TODO: this is temporary solution: for post-aggregate calculations we
    # consider the measure reference to be aggregated measure reference.
    # TODO: this does not work for implicit post-aggregate calculations

    function = WindowFunction(window_function, window_key,
                              target_attribute=aggregate.name,
                              source_attribute=source,
                              window_size=window_size,
                              label=label)
    return function

def get_key(record, composite_key):
    """Extracts a tuple of values from the `record` by `composite_key`"""
    return tuple(record.get(key) for key in composite_key)

class WindowFunction(object):
    def __init__(self, function, window_key, target_attribute,
                 source_attribute, window_size, label):
        """Creates a window function."""

        if not function:
            raise ArgumentError("No window function provided")
        if window_size < 1:
            raise ArgumentError("Window size should be >= 1")
        if not source_attribute:
            raise ArgumentError("Source attribute not specified")
        if not target_attribute:
            raise ArgumentError("Target attribute not specified")

        self.function = function
        self.window_key = tuple(window_key) if window_key else tuple()
        self.source_attribute = source_attribute
        self.target_attribute = target_attribute
        self.window_size = window_size
        self.window_values = {}
        self.label = label

    def __call__(self, record):
        """Collects the source value. If the window for the `window_key` is
        filled, then apply the window function and store the value in the
        `record` to key `target_attribute`."""

        key = get_key(record, self.window_key)

        # Get the window values by key. Create new if necessary.
        try:
            values = self.window_values[key]
        except KeyError:
            values = deque()
            self.window_values[key] = values

        value = record.get(self.source_attribute)

        # TODO: What about those window functions that would want to have empty
        # values?
        if value is not None:
            values.append(value)

        # Keep the window within the window size:
        while len(values) > self.window_size:
            values.popleft()

        # Compute, if we have the values
        if len(values) > 0:
            record[self.target_attribute] = self.function(values)



# TODO: make CALCULATED_AGGREGATIONS a namespace (see extensions.py)
CALCULATED_AGGREGATIONS = {
    "wma": partial(_window_function_factory,
                   window_function=weighted_moving_average,
                   label='Weighted Moving Avg. of {measure}'),
    "sma": partial(_window_function_factory,
                   window_function=simple_moving_average,
                   label='Simple Moving Avg. of {measure}'),
    "sms": partial(_window_function_factory,
                   window_function=simple_moving_sum,
                   label='Simple Moving Sum of {measure}'),
    "smstd": partial(_window_function_factory,
                     window_function=simple_stdev,
                     label='Moving Std. Deviation of {measure}'),
    "smrsd": partial(_window_function_factory,
                     window_function=simple_relative_stdev,
                     label='Moving Relative St. Dev. of {measure}'),
    "smvar": partial(_window_function_factory,
                     window_function=simple_variance,
                     label='Moving Variance of {measure}')
}

def available_calculators():
    """Returns a list of available calculators."""
    return CALCULATED_AGGREGATIONS.keys()

def aggregate_calculator_labels():
    return dict([(k, v.keywords['label']) for k, v in CALCULATED_AGGREGATIONS.items()])
